import asyncio
from asyncio import ensure_future
from functools import wraps
from traceback import format_exception
from typing import Callable, Coroutine, Any

from loguru import logger
from starlette.concurrency import run_in_threadpool

NoArgsNoReturnFuncT = Callable[[], None]
NoArgsNoReturnAsyncFuncT = Callable[[], Coroutine[Any, Any, None]]
NoArgsNoReturnDecorator = Callable[
    [NoArgsNoReturnFuncT | NoArgsNoReturnAsyncFuncT], NoArgsNoReturnAsyncFuncT
]


def repeat_every(
    seconds: float,
    wait_first: bool = False,
    raise_exceptions: bool = False,
    max_repetitions: int | None = None,
) -> NoArgsNoReturnDecorator:
    """
    This function returns a decorator that modifies a function so it is periodically re-executed after its first call.

    The function it decorates should accept no arguments and return nothing. If necessary, this can be accomplished
    by using `functools.partial` or otherwise wrapping the target function prior to decoration.

    Parameters
    ----------
    seconds: float
        The number of seconds to wait between repeated calls
    wait_first: bool (default False)
        If True, the function will wait for a single period before the first call
    raise_exceptions: bool (default False)
        If True, errors raised by the decorated function will be raised to the event loop's exception handler.
        Note that if an error is raised, the repeated execution will stop.
        Otherwise, exceptions are just logged and the execution continues to repeat.
        See https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.set_exception_handler for more info.
    max_repetitions: Optional[int] (default None)
        The maximum number of times to call the repeated function. If `None`, the function is repeated forever.
    """

    def decorator(
        func: NoArgsNoReturnAsyncFuncT | NoArgsNoReturnFuncT,
    ) -> NoArgsNoReturnAsyncFuncT:
        """
        Converts the decorated function into a repeated, periodically-called version of itself.
        """
        is_coroutine = asyncio.iscoroutinefunction(func)

        @wraps(func)
        async def wrapped() -> None:
            repetitions = 0

            async def loop() -> None:
                nonlocal repetitions

                if wait_first:
                    await asyncio.sleep(seconds)
                while max_repetitions is None or repetitions < max_repetitions:
                    # count the repetition even if an exception is raised
                    repetitions += 1
                    try:
                        if is_coroutine:
                            await func()  # type: ignore
                        else:
                            await run_in_threadpool(func)
                    except Exception as exc:
                        formatted_exception = "".join(
                            format_exception(type(exc), exc, exc.__traceback__)
                        )
                        logger.error(formatted_exception)
                        if raise_exceptions:
                            raise exc
                    await asyncio.sleep(seconds)

            ensure_future(loop())

        return wrapped

    return decorator



from port_ocean.core.event_listener.http import (
    HttpEventListenerSettings,
    HttpEventListener,
)
from port_ocean.core.event_listener.kafka import (
    KafkaEventListenerSettings,
    KafkaEventListener,
)
from port_ocean.core.event_listener.polling import (
    PollingEventListenerSettings,
    PollingEventListener,
)

from port_ocean.core.event_listener.once import (
    OnceEventListenerSettings,
    OnceEventListener,
)


EventListenerSettingsType = (
    HttpEventListenerSettings
    | KafkaEventListenerSettings
    | PollingEventListenerSettings
    | OnceEventListenerSettings
)

__all__ = [
    "EventListenerSettingsType",
    "HttpEventListener",
    "HttpEventListenerSettings",
    "KafkaEventListener",
    "KafkaEventListenerSettings",
    "PollingEventListener",
    "PollingEventListenerSettings",
    "OnceEventListener",
    "OnceEventListenerSettings",
]



import typing

import aiostream

if typing.TYPE_CHECKING:
    from asyncio import Semaphore


async def stream_async_iterators_tasks(
    *tasks: typing.AsyncIterable[typing.Any],
) -> typing.AsyncIterable[typing.Any]:
    """
    This function takes a list of async iterators and streams the results of each iterator as they are available.
    By using this function you can combine multiple async iterators into a single stream of results, instead of waiting
    for each iterator to finish before starting the next one.

    Usage:
    ```python
    async def async_iterator1():
        for i in range(10):
            yield i
            await asyncio.sleep(1)

    async def async_iterator2():
        for i in range(10, 20):
            yield i
            await asyncio.sleep(1)

    async def main():
        async for result in stream_async_iterators_tasks([async_iterator1(), async_iterator2()]):
            print(result)
    ```

    Caution - Before using this function, make sure that the third-party API you are calling allows the number of
    concurrent requests you are making. If the API has a rate limit, you may need to adjust the number of concurrent
    requests to avoid hitting the rate limit.

    :param tasks: A list of async iterators
    :return: A stream of results
    """
    if not tasks:
        return

    if len(tasks) == 1:
        async for batch_items in tasks[0]:
            yield batch_items
        return

    combine = aiostream.stream.merge(tasks[0], *tasks[1:])
    async with combine.stream() as streamer:
        async for batch_items in streamer:
            yield batch_items


async def semaphore_async_iterator(
    semaphore: "Semaphore",
    function: typing.Callable[[], typing.AsyncIterator[typing.Any]],
) -> typing.AsyncIterator[typing.Any]:
    """
    Executes an asynchronous iterator function under a semaphore to limit concurrency.

    This function ensures that the provided asynchronous iterator function is executed
    while respecting the concurrency limit imposed by the semaphore. It acquires the
    semaphore before executing the function and releases it after the function completes,
    thus controlling the number of concurrent executions.

    Parameters:
        semaphore (asyncio.Semaphore | asyncio.BoundedSemaphore): The semaphore used to limit concurrency.
        function (Callable[[], AsyncIterator[Any]]): A nullary asynchronous function, - apply arguments with `functools.partial` or an anonymous function (lambda)
            that returns an asynchronous iterator. This function is executed under the semaphore.

    Yields:
        Any: The items yielded by the asynchronous iterator function.

    Usage:
        ```python
        import asyncio

        async def async_iterator_function(param1, param2):
            # Your async code here
            yield ...

        async def async_generator_function():
            # Your async code to retrieve items
            param1 = "your_param1"
            yield param1

        async def main():
            semaphore = asyncio.BoundedSemaphore(50)
            param2 = "your_param2"

            tasks = [
                semaphore_async_iterator(
                    semaphore,
                    lambda: async_iterator_function(param1, param2) # functools.partial(async_iterator_function, param1, param2)
                )
                async for param1 in async_generator_function()
            ]

            async for batch in stream_async_iterators_tasks(*tasks):
                # Process each batch
                pass

        asyncio.run(main())
        ```
    """
    async with semaphore:
        async for result in function():
            yield result



import asyncio
from typing import Any
from urllib.parse import quote_plus

import httpx
from loguru import logger

from port_ocean.clients.port.authentication import PortAuthentication
from port_ocean.clients.port.types import RequestOptions, UserAgentType
from port_ocean.clients.port.utils import (
    handle_status_code,
    PORT_HTTP_MAX_CONNECTIONS_LIMIT,
)
from port_ocean.core.models import Entity


class EntityClientMixin:
    def __init__(self, auth: PortAuthentication, client: httpx.AsyncClient):
        self.auth = auth
        self.client = client
        # Semaphore is used to limit the number of concurrent requests to port, to avoid overloading it.
        # The number of concurrent requests is set to 90% of the max connections limit, to leave some room for other
        # requests that are not related to entities.
        self.semaphore = asyncio.Semaphore(round(0.9 * PORT_HTTP_MAX_CONNECTIONS_LIMIT))

    async def upsert_entity(
        self,
        entity: Entity,
        request_options: RequestOptions,
        user_agent_type: UserAgentType | None = None,
        should_raise: bool = True,
    ) -> Entity | None:
        validation_only = request_options["validation_only"]
        async with self.semaphore:
            logger.debug(
                f"{'Validating' if validation_only else 'Upserting'} entity: {entity.identifier} of blueprint: {entity.blueprint}"
            )
            headers = await self.auth.headers(user_agent_type)
            response = await self.client.post(
                f"{self.auth.api_url}/blueprints/{entity.blueprint}/entities",
                json=entity.dict(exclude_unset=True, by_alias=True),
                headers=headers,
                params={
                    "upsert": "true",
                    "merge": str(request_options["merge"]).lower(),
                    "create_missing_related_entities": str(
                        request_options["create_missing_related_entities"]
                    ).lower(),
                    "validation_only": str(validation_only).lower(),
                },
            )

        if response.is_error:
            logger.error(
                f"Error {'Validating' if validation_only else 'Upserting'} "
                f"entity: {entity.identifier} of "
                f"blueprint: {entity.blueprint}"
            )
        handle_status_code(response, should_raise)
        result = response.json()

        result_entity = (
            Entity.parse_obj(result["entity"]) if result.get("entity") else entity
        )

        # Happens when upsert fails and search identifier is defined.
        # We return None to ignore the entity later in the delete process
        if result_entity.is_using_search_identifier:
            return None

        # In order to save memory we'll keep only the identifier, blueprint and relations of the
        # upserted entity result for later calculations
        reduced_entity = Entity(
            identifier=result_entity.identifier, blueprint=result_entity.blueprint
        )

        # Turning dict typed relations (raw search relations) is required
        # for us to be able to successfully calculate the participation related entities
        # and ignore the ones that don't as they weren't upserted
        reduced_entity.relations = {
            key: None if isinstance(relation, dict) else relation
            for key, relation in result_entity.relations.items()
        }

        return reduced_entity

    async def batch_upsert_entities(
        self,
        entities: list[Entity],
        request_options: RequestOptions,
        user_agent_type: UserAgentType | None = None,
        should_raise: bool = True,
    ) -> list[Entity]:
        modified_entities_results = await asyncio.gather(
            *(
                self.upsert_entity(
                    entity,
                    request_options,
                    user_agent_type,
                    should_raise=should_raise,
                )
                for entity in entities
            ),
            return_exceptions=True,
        )
        entity_results = [
            entity for entity in modified_entities_results if isinstance(entity, Entity)
        ]
        if not should_raise:
            return entity_results

        for entity_result in modified_entities_results:
            if isinstance(entity_result, Exception):
                raise entity_result

        return entity_results

    async def delete_entity(
        self,
        entity: Entity,
        request_options: RequestOptions,
        user_agent_type: UserAgentType | None = None,
        should_raise: bool = True,
    ) -> None:
        async with self.semaphore:
            logger.info(
                f"Delete entity: {entity.identifier} of blueprint: {entity.blueprint}"
            )
            response = await self.client.delete(
                f"{self.auth.api_url}/blueprints/{entity.blueprint}/entities/{quote_plus(entity.identifier)}",
                headers=await self.auth.headers(user_agent_type),
                params={
                    "delete_dependents": str(
                        request_options["delete_dependent_entities"]
                    ).lower()
                },
            )

            if response.is_error:
                if response.status_code == 404:
                    logger.info(
                        f"Failed to delete entity: {entity.identifier} of blueprint: {entity.blueprint},"
                        f" as it was already deleted from port"
                    )
                    return
                logger.error(
                    f"Error deleting "
                    f"entity: {entity.identifier} of "
                    f"blueprint: {entity.blueprint}"
                )

            handle_status_code(response, should_raise)

    async def batch_delete_entities(
        self,
        entities: list[Entity],
        request_options: RequestOptions,
        user_agent_type: UserAgentType | None = None,
        should_raise: bool = True,
    ) -> None:
        await asyncio.gather(
            *(
                self.delete_entity(
                    entity,
                    request_options,
                    user_agent_type,
                    should_raise=should_raise,
                )
                for entity in entities
            ),
            return_exceptions=True,
        )

    async def search_entities(
        self, user_agent_type: UserAgentType, query: dict[Any, Any] | None = None
    ) -> list[Entity]:
        default_query = {
            "combinator": "and",
            "rules": [
                {
                    "property": "$datasource",
                    "operator": "contains",
                    "value": f"port-ocean/{self.auth.integration_type}/",
                },
                {
                    "property": "$datasource",
                    "operator": "contains",
                    "value": f"/{self.auth.integration_identifier}/{user_agent_type.value}",
                },
            ],
        }

        if query is None:
            query = default_query
        elif query.get("rules"):
            query["rules"].append(default_query)

        logger.info(f"Searching entities with query {query}")
        response = await self.client.post(
            f"{self.auth.api_url}/entities/search",
            json=query,
            headers=await self.auth.headers(user_agent_type),
            params={
                "exclude_calculated_properties": "true",
                "include": ["blueprint", "identifier"],
            },
            extensions={"retryable": True},
            timeout=30,
        )
        handle_status_code(response)
        return [Entity.parse_obj(result) for result in response.json()["entities"]]

    async def search_batch_entities(
        self, user_agent_type: UserAgentType, entities_to_search: list[Entity]
    ) -> list[Entity]:
        search_rules = []
        for entity in entities_to_search:
            search_rules.append(
                {
                    "combinator": "and",
                    "rules": [
                        {
                            "property": "$identifier",
                            "operator": "=",
                            "value": entity.identifier,
                        },
                        {
                            "property": "$blueprint",
                            "operator": "=",
                            "value": entity.blueprint,
                        },
                    ],
                }
            )

        return await self.search_entities(
            user_agent_type,
            {
                "combinator": "and",
                "rules": [{"combinator": "or", "rules": search_rules}],
            },
        )



from asyncio import Task, get_event_loop
from typing import Literal, Any

from loguru import logger

from port_ocean.context.ocean import ocean
from port_ocean.core.event_listener.base import (
    BaseEventListener,
    EventListenerEvents,
    EventListenerSettings,
)
from port_ocean.utils.repeat import repeat_every
from port_ocean.utils.signal import signal_handler


class PollingEventListenerSettings(EventListenerSettings):
    """
    Attributes:
        type (Literal["POLLING"]): A literal indicating the type of the event listener, which is set to "POLLING" for this class.
        resync_on_start (bool): A flag indicating whether to trigger a resync event on the start of the polling event listener.
                                If True, the "on_resync" event will be triggered immediately when the polling listener starts.
        interval (int): The interval in seconds at which the polling event listener checks for changes in the integration.
                        The default interval is set to 60 seconds.
    """

    type: Literal["POLLING"]
    resync_on_start: bool = True
    interval: int = 60

    def to_request(self) -> dict[str, Any]:
        return {}


class PollingEventListener(BaseEventListener):
    """
    Polling event listener that checks for changes in the integration every `interval` seconds.

    The `PollingEventListener` periodically checks for changes in the integration and triggers the "on_resync" event if changes are detected.

    Parameters:
        events (EventListenerEvents): A dictionary containing event types and their corresponding event handlers.
        event_listener_config (PollingEventListenerSettings): Configuration settings for the Polling event listener.
    """

    def __init__(
        self,
        events: EventListenerEvents,
        event_listener_config: PollingEventListenerSettings,
    ):
        super().__init__(events)
        self.event_listener_config = event_listener_config

    def should_resync(self, last_updated_at: str) -> bool:
        _last_updated_at = (
            ocean.app.resync_state_updater.last_integration_state_updated_at
        )

        if _last_updated_at is None:
            return self.event_listener_config.resync_on_start

        return _last_updated_at != last_updated_at

    async def _start(self) -> None:
        """
        Starts the polling event listener.
        It registers the "on_resync" event to be called every `interval` seconds specified in the `event_listener_config`.
        The `on_resync` event is triggered if the integration has changed since the last update.
        """
        logger.info(
            f"Setting up Polling event listener with interval: {self.event_listener_config.interval}"
        )

        @repeat_every(seconds=self.event_listener_config.interval)
        async def resync() -> None:
            logger.info(
                f"Polling event listener iteration after {self.event_listener_config.interval}. Checking for changes"
            )
            integration = await ocean.app.port_client.get_current_integration()
            last_updated_at = integration["updatedAt"]

            if self.should_resync(last_updated_at):
                logger.info("Detected change in integration, resyncing")
                ocean.app.resync_state_updater.last_integration_state_updated_at = (
                    last_updated_at
                )
                running_task: Task[Any] = get_event_loop().create_task(self._resync({}))
                signal_handler.register(running_task.cancel)

                await running_task

        # Execute resync repeatedly task
        await resync()



import asyncio
import inspect
import typing
from typing import Callable, Awaitable, Any

import httpx
from loguru import logger

from port_ocean.clients.port.types import UserAgentType
from port_ocean.context.event import TriggerType, event_context, EventType, event
from port_ocean.context.ocean import ocean
from port_ocean.context.resource import resource_context
from port_ocean.core.handlers.port_app_config.models import ResourceConfig
from port_ocean.core.integrations.mixins import HandlerMixin, EventsMixin
from port_ocean.core.integrations.mixins.utils import (
    is_resource_supported,
    unsupported_kind_response,
    resync_generator_wrapper,
    resync_function_wrapper,
)
from port_ocean.core.models import Entity
from port_ocean.core.ocean_types import (
    RAW_RESULT,
    RESYNC_RESULT,
    RawEntityDiff,
    ASYNC_GENERATOR_RESYNC_TYPE,
    RAW_ITEM,
    CalculationResult,
)
from port_ocean.core.utils import zip_and_sum, gather_and_split_errors_from_results
from port_ocean.exceptions.core import OceanAbortException

SEND_RAW_DATA_EXAMPLES_AMOUNT = 5


class SyncRawMixin(HandlerMixin, EventsMixin):
    """Mixin class for synchronization of raw constructed entities.

    This mixin class extends the functionality of HandlerMixin and EventsMixin to provide methods for registering,
    unregistering, updating, and syncing raw entities' state changes.

    Note:
        Raw entities are entities with a more primitive structure, usually fetched directly from a resource.
    """

    def __init__(self) -> None:
        HandlerMixin.__init__(self)
        EventsMixin.__init__(self)

    async def _on_resync(self, kind: str) -> RAW_RESULT:
        raise NotImplementedError("on_resync must be implemented")

    async def _get_resource_raw_results(
        self, resource_config: ResourceConfig
    ) -> tuple[RESYNC_RESULT, list[Exception]]:
        logger.info(f"Fetching {resource_config.kind} resync results")

        if not is_resource_supported(
            resource_config.kind, self.event_strategy["resync"]
        ):
            return unsupported_kind_response(
                resource_config.kind, self.available_resync_kinds
            )

        fns = self._collect_resync_functions(resource_config)

        results, errors = await self._execute_resync_tasks(fns, resource_config)

        return results, errors

    def _collect_resync_functions(
        self, resource_config: ResourceConfig
    ) -> list[Callable[[str], Awaitable[RAW_RESULT]]]:
        logger.contextualize(kind=resource_config.kind)

        fns = [
            *self.event_strategy["resync"][resource_config.kind],
            *self.event_strategy["resync"][None],
        ]

        if self.__class__._on_resync != SyncRawMixin._on_resync:
            fns.append(self._on_resync)

        return fns

    async def _execute_resync_tasks(
        self,
        fns: list[Callable[[str], Awaitable[RAW_RESULT]]],
        resource_config: ResourceConfig,
    ) -> tuple[RESYNC_RESULT, list[RAW_RESULT | Exception]]:
        tasks = []
        results = []
        for task in fns:
            if inspect.isasyncgenfunction(task):
                results.append(resync_generator_wrapper(task, resource_config.kind))
            else:
                task = typing.cast(Callable[[str], Awaitable[RAW_RESULT]], task)
                tasks.append(resync_function_wrapper(task, resource_config.kind))

        logger.info(
            f"Found {len(tasks) + len(results)} resync tasks for {resource_config.kind}"
        )
        successful_results, errors = await gather_and_split_errors_from_results(tasks)
        results.extend(
            sum(
                successful_results,
                [],
            )
        )

        logger.info(
            f"Triggered {len(tasks)} tasks for {resource_config.kind}, failed: {len(errors)}"
        )

        return results, errors

    async def _calculate_raw(
        self,
        raw_diff: list[tuple[ResourceConfig, list[RAW_ITEM]]],
        parse_all: bool = False,
        send_raw_data_examples_amount: int = 0,
    ) -> list[CalculationResult]:
        return await asyncio.gather(
            *(
                self.entity_processor.parse_items(
                    mapping, results, parse_all, send_raw_data_examples_amount
                )
                for mapping, results in raw_diff
            )
        )

    async def _register_resource_raw(
        self,
        resource: ResourceConfig,
        results: list[dict[Any, Any]],
        user_agent_type: UserAgentType,
        parse_all: bool = False,
        send_raw_data_examples_amount: int = 0,
    ) -> CalculationResult:
        objects_diff = await self._calculate_raw(
            [(resource, results)], parse_all, send_raw_data_examples_amount
        )
        modified_objects = await self.entities_state_applier.upsert(
            objects_diff[0].entity_selector_diff.passed, user_agent_type
        )
        return CalculationResult(
            objects_diff[0].entity_selector_diff._replace(passed=modified_objects),
            errors=objects_diff[0].errors,
        )

    async def _unregister_resource_raw(
        self,
        resource: ResourceConfig,
        results: list[RAW_ITEM],
        user_agent_type: UserAgentType,
    ) -> tuple[list[Entity], list[Exception]]:
        if resource.port.entity.mappings.is_using_search_identifier:
            logger.info(
                f"Skip unregistering resource of kind {resource.kind}, as mapping defined with search identifier"
            )
            return [], []

        objects_diff = await self._calculate_raw([(resource, results)])
        entities_selector_diff, errors = objects_diff[0]

        await self.entities_state_applier.delete(
            entities_selector_diff.passed, user_agent_type
        )
        logger.info("Finished unregistering change")
        return entities_selector_diff.passed, errors

    async def _register_in_batches(
        self, resource_config: ResourceConfig, user_agent_type: UserAgentType
    ) -> tuple[list[Entity], list[Exception]]:
        results, errors = await self._get_resource_raw_results(resource_config)
        async_generators: list[ASYNC_GENERATOR_RESYNC_TYPE] = []
        raw_results: RAW_RESULT = []
        for result in results:
            if isinstance(result, dict):
                raw_results.append(result)
            else:
                async_generators.append(result)

        send_raw_data_examples_amount = (
            SEND_RAW_DATA_EXAMPLES_AMOUNT if ocean.config.send_raw_data_examples else 0
        )
        all_entities, register_errors = await self._register_resource_raw(
            resource_config,
            raw_results,
            user_agent_type,
            send_raw_data_examples_amount=send_raw_data_examples_amount,
        )
        errors.extend(register_errors)
        passed_entities = list(all_entities.passed)

        for generator in async_generators:
            try:
                async for items in generator:
                    if send_raw_data_examples_amount > 0:
                        send_raw_data_examples_amount = max(
                            0, send_raw_data_examples_amount - len(passed_entities)
                        )

                    entities, register_errors = await self._register_resource_raw(
                        resource_config,
                        items,
                        user_agent_type,
                        send_raw_data_examples_amount=send_raw_data_examples_amount,
                    )
                    errors.extend(register_errors)
                    passed_entities.extend(entities.passed)
            except* OceanAbortException as error:
                errors.append(error)

        logger.info(
            f"Finished registering change for {len(results)} raw results for kind: {resource_config.kind}. {len(passed_entities)} entities were affected"
        )
        return passed_entities, errors

    async def register_raw(
        self,
        kind: str,
        results: list[dict[Any, Any]],
        user_agent_type: UserAgentType,
    ) -> list[Entity]:
        """Register raw entities of a specific kind.

        This method registers raw entities of a specific kind into Port.

        Args:
            kind (str): The kind of raw entities being registered.
            results (list[dict[Any, Any]]): The raw entity results to be registered.
            user_agent_type (UserAgentType): The type of user agent.

        Returns:
            list[Entity]: A list of registered entities.
        """
        logger.info(f"Registering state for {kind}")
        config = await self.port_app_config_handler.get_port_app_config()
        resource_mappings = [
            resource for resource in config.resources if resource.kind == kind
        ]

        if not resource_mappings:
            return []

        diffs, errors = zip(
            *await asyncio.gather(
                *(
                    self._register_resource_raw(
                        resource, results, user_agent_type, True
                    )
                    for resource in resource_mappings
                )
            )
        )

        diffs = list(diffs)
        errors = sum(errors, [])

        if errors:
            message = f"Failed to register {len(errors)} entities. Skipping delete phase due to incomplete state"
            logger.error(message, exc_info=errors)
            raise ExceptionGroup(
                message,
                errors,
            )

        registered_entities, entities_to_delete = zip_and_sum(diffs)

        registered_entities_attributes = {
            (entity.identifier, entity.blueprint) for entity in registered_entities
        }

        filtered_entities_to_delete: list[Entity] = (
            await ocean.port_client.search_batch_entities(
                user_agent_type,
                [
                    entity
                    for entity in entities_to_delete
                    if not entity.is_using_search_identifier
                    and (entity.identifier, entity.blueprint)
                    not in registered_entities_attributes
                ],
            )
        )

        if filtered_entities_to_delete:
            logger.info(
                f"Deleting {len(filtered_entities_to_delete)} entities that didn't pass any of the selectors"
            )

            await self.entities_state_applier.delete(
                filtered_entities_to_delete, user_agent_type
            )

        return registered_entities

    async def unregister_raw(
        self,
        kind: str,
        results: list[dict[Any, Any]],
        user_agent_type: UserAgentType,
    ) -> list[Entity]:
        """Unregister raw entities of a specific kind.

        This method unregisters raw entities of a specific kind from Port.

        Args:
            kind (str): The kind of raw entities being unregistered.
            results (list[dict[Any, Any]]): The raw entity results to be unregistered.
            user_agent_type (UserAgentType): The type of user agent.

        Returns:
            list[Entity]: A list of unregistered entities.
        """
        logger.info(f"Registering state for {kind}")
        config = await self.port_app_config_handler.get_port_app_config()
        resource_mappings = [
            resource for resource in config.resources if resource.kind == kind
        ]

        entities, errors = zip_and_sum(
            await asyncio.gather(
                *(
                    self._unregister_resource_raw(resource, results, user_agent_type)
                    for resource in resource_mappings
                )
            )
        )

        if errors:
            message = f"Failed to unregister all entities with {len(errors)} errors"
            logger.error(message, exc_info=errors)
            raise ExceptionGroup(
                message,
                errors,
            )

        return entities

    async def update_raw_diff(
        self,
        kind: str,
        raw_desired_state: RawEntityDiff,
        user_agent_type: UserAgentType,
    ) -> None:
        """Update the difference in state for raw entities of a specific kind.

        This method updates the difference in state for raw entities of a specific kind.

        Args:
            kind (str): The kind of raw entities being updated.
            raw_desired_state (RawEntityDiff): The desired state difference of raw entities.
            user_agent_type (UserAgentType): The type of user agent.
        """
        logger.info(f"Updating state for {kind}")
        config = await self.port_app_config_handler.get_port_app_config()
        resource_mappings = [
            resource for resource in config.resources if resource.kind == kind
        ]

        with logger.contextualize(kind=kind):
            logger.info(f"Found {len(resource_mappings)} resources for {kind}")

            entities_before, _ = zip(
                await self._calculate_raw(
                    [
                        (mapping, raw_desired_state["before"])
                        for mapping in resource_mappings
                    ]
                )
            )

            entities_after, after_errors = await self._calculate_raw(
                [(mapping, raw_desired_state["after"]) for mapping in resource_mappings]
            )

            entities_before_flatten: list[Entity] = sum(
                (entities_diff.passed for entities_diff in entities_before), []
            )

            entities_after_flatten: list[Entity] = sum(
                (entities_diff.passed for entities_diff in entities_after), []
            )

            if after_errors:
                message = (
                    f"Failed to calculate diff for entities with {len(after_errors)} errors. "
                    f"Skipping delete phase due to incomplete state"
                )
                logger.error(message, exc_info=after_errors)
                entities_before_flatten = []

            await self.entities_state_applier.apply_diff(
                {"before": entities_before_flatten, "after": entities_after_flatten},
                user_agent_type,
            )

    async def sync_raw_all(
        self,
        _: dict[Any, Any] | None = None,
        trigger_type: TriggerType = "machine",
        user_agent_type: UserAgentType = UserAgentType.exporter,
        silent: bool = True,
    ) -> None:
        """Perform a full synchronization of raw entities.

        This method performs a full synchronization of raw entities, including registration, unregistration,
        and state updates.

        Args:
            _ (dict[Any, Any] | None): Unused parameter.
            trigger_type (TriggerType): The type of trigger for the synchronization.
            user_agent_type (UserAgentType): The type of user agent.
            silent (bool): Whether to raise exceptions or handle them silently.
        """
        logger.info("Resync was triggered")
        async with event_context(
            EventType.RESYNC,
            trigger_type=trigger_type,
        ):
            # If a resync is triggered due to a mappings change, we want to make sure that we have the updated version
            # rather than the old cache
            app_config = await self.port_app_config_handler.get_port_app_config(
                use_cache=False
            )
            logger.info(f"Resync will use the following mappings: {app_config.dict()}")
            try:
                did_fetched_current_state = True
                entities_at_port = await ocean.port_client.search_entities(
                    user_agent_type
                )
            except httpx.HTTPError as e:
                logger.warning(
                    "Failed to fetch the current state of entities at Port. "
                    "Skipping delete phase due to unknown initial state. "
                    f"Error: {e}\n"
                    f"Response status code: {e.response.status_code if isinstance(e, httpx.HTTPStatusError) else None}\n"
                    f"Response content: {e.response.text if isinstance(e, httpx.HTTPStatusError) else None}\n"
                )
                did_fetched_current_state = False

            creation_results: list[tuple[list[Entity], list[Exception]]] = []

            try:
                for resource in app_config.resources:
                    # create resource context per resource kind, so resync method could have access to the resource
                    # config as we might have multiple resources in the same event
                    async with resource_context(resource):
                        task = asyncio.get_event_loop().create_task(
                            self._register_in_batches(resource, user_agent_type)
                        )

                        event.on_abort(lambda: task.cancel())

                        creation_results.append(await task)
            except asyncio.CancelledError as e:
                logger.warning("Resync aborted successfully, skipping delete phase. This leads to an incomplete state")
                raise
            else:
                if not did_fetched_current_state:
                    logger.warning(
                        "Due to an error before the resync, the previous state of entities at Port is unknown."
                        " Skipping delete phase due to unknown initial state."
                    )
                    return

                logger.info("Starting resync diff calculation")
                flat_created_entities, errors = zip_and_sum(creation_results) or [
                    [],
                    [],
                ]

                if errors:
                    message = f"Resync failed with {len(errors)}. Skipping delete phase due to incomplete state"
                    error_group = ExceptionGroup(
                        f"Resync failed with {len(errors)}. Skipping delete phase due to incomplete state",
                        errors,
                    )
                    if not silent:
                        raise error_group

                    logger.error(message, exc_info=error_group)
                else:
                    logger.info(
                        f"Running resync diff calculation, number of entities at Port before resync: {len(entities_at_port)}, number of entities created during sync: {len(flat_created_entities)}"
                    )
                    await self.entities_state_applier.delete_diff(
                        {"before": entities_at_port, "after": flat_created_entities},
                        user_agent_type,
                    )
                    logger.info("Resync finished successfully")



import datetime
from typing import Any, Literal
from port_ocean.clients.port.client import PortClient
from port_ocean.utils.misc import IntegrationStateStatus
from port_ocean.utils.time import get_next_occurrence


class ResyncStateUpdater:
    def __init__(self, port_client: PortClient, scheduled_resync_interval: int | None):
        self.port_client = port_client
        self.initiated_at = datetime.datetime.now(tz=datetime.timezone.utc)
        self.scheduled_resync_interval = scheduled_resync_interval

        # This is used to differ between integration changes that require a full resync and state changes
        # So that the polling event-listener can decide whether to perform a full resync or not
        # TODO: remove this once we separate the state from the integration
        self.last_integration_state_updated_at: str = ""

    def _calculate_next_scheduled_resync(
        self,
        interval: int | None = None,
        custom_start_time: datetime.datetime | None = None,
    ) -> str | None:
        if interval is None:
            return None
        return get_next_occurrence(
            interval * 60, custom_start_time or self.initiated_at
        ).isoformat()

    async def update_before_resync(
        self,
        interval: int | None = None,
        custom_start_time: datetime.datetime | None = None,
    ) -> None:
        _interval = interval or self.scheduled_resync_interval
        nest_resync = self._calculate_next_scheduled_resync(
            _interval, custom_start_time
        )
        state: dict[str, Any] = {
            "status": IntegrationStateStatus.Running.value,
            "lastResyncEnd": None,
            "lastResyncStart": datetime.datetime.now(
                tz=datetime.timezone.utc
            ).isoformat(),
            "nextResync": nest_resync,
            "intervalInMinuets": _interval,
        }

        integration = await self.port_client.update_integration_state(
            state, should_raise=False
        )
        if integration:
            self.last_integration_state_updated_at = integration["resyncState"][
                "updatedAt"
            ]

    async def update_after_resync(
        self,
        status: Literal[
            IntegrationStateStatus.Completed, IntegrationStateStatus.Failed
        ] = IntegrationStateStatus.Completed,
        interval: int | None = None,
        custom_start_time: datetime.datetime | None = None,
    ) -> None:
        _interval = interval or self.scheduled_resync_interval
        nest_resync = self._calculate_next_scheduled_resync(
            _interval, custom_start_time
        )
        state: dict[str, Any] = {
            "status": status.value,
            "lastResyncEnd": datetime.datetime.now(
                tz=datetime.timezone.utc
            ).isoformat(),
            "nextResync": nest_resync,
            "intervalInMinuets": _interval,
        }

        integration = await self.port_client.update_integration_state(
            state, should_raise=False
        )
        if integration:
            self.last_integration_state_updated_at = integration["resyncState"][
                "updatedAt"
            ]



from dataclasses import dataclass, field
from enum import Enum
from typing import Any

from pydantic import BaseModel
from pydantic.fields import Field


class Runtime(Enum):
    Saas = "Saas"
    OnPrem = "OnPrem"


class Entity(BaseModel):
    identifier: Any
    blueprint: Any
    title: Any
    team: str | None | list[Any] = []
    properties: dict[str, Any] = {}
    relations: dict[str, Any] = {}

    @property
    def is_using_search_identifier(self) -> bool:
        return isinstance(self.identifier, dict)


class BlueprintRelation(BaseModel):
    many: bool
    required: bool
    target: str
    title: str | None


class Blueprint(BaseModel):
    identifier: str
    title: str | None
    team: str | None
    properties_schema: dict[str, Any] = Field(alias="schema")
    relations: dict[str, BlueprintRelation]


class Migration(BaseModel):
    id: str
    actor: str
    sourceBlueprint: str
    mapping: dict[str, Any]
    status: str


@dataclass
class EntityPortDiff:
    """Represents the differences between entities for porting.

    This class holds the lists of deleted, modified, and created entities as part
    of the porting process.
    """

    deleted: list[Entity] = field(default_factory=list)
    modified: list[Entity] = field(default_factory=list)
    created: list[Entity] = field(default_factory=list)



import asyncio
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from typing import (
    AsyncIterator,
    Literal,
    Any,
    TYPE_CHECKING,
    Optional,
    Callable,
    Awaitable,
    Union,
)
from uuid import uuid4

from loguru import logger
from pydispatch import dispatcher  # type: ignore
from werkzeug.local import LocalStack, LocalProxy

from port_ocean.context.resource import resource
from port_ocean.exceptions.context import (
    EventContextNotFoundError,
    ResourceContextNotFoundError,
)
from port_ocean.utils.misc import get_time

if TYPE_CHECKING:
    from port_ocean.core.handlers.port_app_config.models import (
        ResourceConfig,
        PortAppConfig,
    )

TriggerType = Literal["manual", "machine", "request"]
AbortCallbackFunction = Callable[[], Union[Any, Awaitable[Any]]]


class EventType:
    START = "start"
    RESYNC = "resync"
    HTTP_REQUEST = "http_request"


@dataclass
class EventContext:
    event_type: str
    trigger_type: TriggerType = "machine"
    attributes: dict[str, Any] = field(default_factory=dict)
    _aborted: bool = False
    _port_app_config: Optional["PortAppConfig"] = None
    _parent_event: Optional["EventContext"] = None
    _event_id: str = field(default_factory=lambda: str(uuid4()))
    _on_abort_callbacks: list[AbortCallbackFunction] = field(default_factory=list)

    def on_abort(self, func: AbortCallbackFunction) -> None:
        self._on_abort_callbacks.append(func)

    def abort(self) -> None:
        for func in self._on_abort_callbacks:
            try:
                if asyncio.iscoroutinefunction(func):
                    asyncio.get_running_loop().run_until_complete(func())
                else:
                    func()
            except Exception as ex:
                logger.warning(
                    f"Failed to call one of the abort callbacks {ex}", exc_info=True
                )
        self._aborted = True

    @property
    def aborted(self) -> bool:
        return self._aborted

    @property
    def resource_config(self) -> Optional["ResourceConfig"]:
        try:
            return resource.resource_config
        except ResourceContextNotFoundError:
            return None

    @property
    def id(self) -> str:
        return self._event_id

    @property
    def parent(self) -> Optional["EventContext"]:
        return self._parent_event

    @property
    def parent_id(self) -> Optional[str]:
        return self._parent_event.id if self._parent_event else None

    @property
    def port_app_config(self) -> "PortAppConfig":
        if self._port_app_config is None:
            raise ValueError("Port app config is not set")
        return self._port_app_config

    @port_app_config.setter
    def port_app_config(self, value: "PortAppConfig") -> None:
        self._port_app_config = value


_event_context_stack: LocalStack[EventContext] = LocalStack()


def _get_event_context() -> EventContext:
    """
    Get the event context from the current thread.
    """
    top_event_context = _event_context_stack.top
    if top_event_context is None:
        raise EventContextNotFoundError(
            "You must be inside an event context in order to use it"
        )

    return top_event_context


event: EventContext = LocalProxy(lambda: _get_event_context())  # type: ignore


@asynccontextmanager
async def event_context(
    event_type: str,
    trigger_type: TriggerType = "manual",
    attributes: dict[str, Any] | None = None,
    parent_override: EventContext | None = None,
) -> AsyncIterator[EventContext]:
    parent = parent_override or _event_context_stack.top
    parent_attributes = parent.attributes if parent else {}

    attributes = {**parent_attributes, **(attributes or {})}
    new_event = EventContext(
        event_type,
        trigger_type=trigger_type,
        attributes=attributes,
        _parent_event=parent,
        # inherit port app config from parent event, so it can be used in nested events
        _port_app_config=parent.port_app_config if parent else None,
    )
    _event_context_stack.push(new_event)

    def _handle_event(triggering_event_id: int) -> None:
        if (
            new_event.event_type == EventType.RESYNC
            and new_event.id != triggering_event_id
        ):
            logger.warning("ABORTING RESYNC EVENT DUE TO NEWER RESYNC REQUEST")
            new_event.abort()

    dispatcher.connect(_handle_event, event_type)
    dispatcher.send(event_type, triggering_event_id=event.id)

    start_time = get_time(seconds_precision=False)
    with logger.contextualize(
        event_trigger_type=event.trigger_type,
        event_kind=event.event_type,
        event_id=event.id,
        event_parent_id=event.parent_id,
        event_resource_kind=(
            event.resource_config.kind if event.resource_config else None
        ),
    ):
        logger.info("Event started")
        try:
            yield event
        except:
            success = False
            raise
        else:
            success = True
        finally:
            end_time = get_time(seconds_precision=False)
            time_elapsed = round(end_time - start_time, 5)
            logger.bind(
                success=success,
                time_elapsed=time_elapsed,
            ).info("Event finished")

            dispatcher.disconnect(_handle_event, event_type)

    _event_context_stack.pop()



import asyncio
import logging
import sys
import threading
import time
from datetime import datetime
from logging.handlers import MemoryHandler
from typing import Any

from loguru import logger

from port_ocean import Ocean
from port_ocean.context.ocean import ocean
from copy import deepcopy
from traceback import format_exception


def _serialize_record(record: logging.LogRecord) -> dict[str, Any]:
    extra = {**deepcopy(record.__dict__["extra"])}
    if isinstance(extra.get("exc_info"), Exception):
        serialized_exception = "".join(format_exception(extra.get("exc_info")))
        extra["exc_info"] = serialized_exception
    return {
        "message": record.msg,
        "level": record.levelname,
        "timestamp": datetime.utcfromtimestamp(record.created).strftime(
            "%Y-%m-%dT%H:%M:%S.%fZ"
        ),
        "extra": extra,
    }


class HTTPMemoryHandler(MemoryHandler):
    def __init__(
        self,
        capacity: int = 100,
        flush_level: int = logging.FATAL,
        flush_interval: int = 5,
        flush_size: int = 1024,
    ):
        super().__init__(capacity, flushLevel=flush_level, target=None)
        self.flush_interval = flush_interval
        self.flush_size = flush_size
        self.last_flush_time = time.time()
        self._serialized_buffer: list[dict[str, Any]] = []
        self._thread_pool: list[threading.Thread] = []

    @property
    def ocean(self) -> Ocean | None:
        # We want to wait for the context to be initialized before we can send logs
        if ocean.initialized:
            return ocean.app
        return None

    def emit(self, record: logging.LogRecord) -> None:

        self._serialized_buffer.append(_serialize_record(record))
        super().emit(record)

    def shouldFlush(self, record: logging.LogRecord) -> bool:
        """
        Extending shouldFlush to include size and time validation as part of the decision whether to flush
        """
        if bool(self.buffer) and (
            super(HTTPMemoryHandler, self).shouldFlush(record)
            or sys.getsizeof(self.buffer) >= self.flush_size
            or time.time() - self.last_flush_time >= self.flush_interval
        ):
            return True
        return False

    def wait_for_lingering_threads(self) -> None:
        for thread in self._thread_pool:
            if thread.is_alive():
                thread.join()

    def flush(self) -> None:
        if self.ocean is None or not self.buffer:
            return

        def _wrap_event_loop(_ocean: Ocean, logs_to_send: list[dict[str, Any]]) -> None:
            loop = asyncio.new_event_loop()
            loop.run_until_complete(self.send_logs(_ocean, logs_to_send))
            loop.close()

        def clear_thread_pool() -> None:
            for thread in self._thread_pool:
                if not thread.is_alive():
                    self._thread_pool.remove(thread)

        self.acquire()
        logs = list(self._serialized_buffer)
        if logs:
            self.buffer.clear()
            self._serialized_buffer.clear()
            self.last_flush_time = time.time()
            clear_thread_pool()
            thread = threading.Thread(target=_wrap_event_loop, args=(self.ocean, logs))
            thread.start()
            self._thread_pool.append(thread)
        self.release()

    async def send_logs(
        self, _ocean: Ocean, logs_to_send: list[dict[str, Any]]
    ) -> None:
        try:
            await _ocean.port_client.ingest_integration_logs(logs_to_send)
        except Exception as e:
            logger.debug(f"Failed to send logs to Port with error: {e}")



from typing import Literal, Any

from fastapi import APIRouter
from loguru import logger
from pydantic import AnyHttpUrl
from pydantic.fields import Field

from port_ocean.context.ocean import ocean
from port_ocean.core.event_listener.base import (
    BaseEventListener,
    EventListenerEvents,
    EventListenerSettings,
)


class HttpEventListenerSettings(EventListenerSettings):
    """
    This class inherits from `EventListenerSettings`, which provides a foundation for defining event listener configurations.
    The `HttpEventListenerSettings` specifically includes settings related to the HTTP event listener (Webhook).

    Attributes:
        type (Literal["WEBHOOK"]): A literal indicating the type of the event listener, which is set to "WEBHOOK" for this class.
        app_host (AnyHttpUrl): The base URL of the application hosting the webhook.
                               The "AnyHttpUrl" type indicates that the value must be a valid HTTP/HTTPS URL.
    """

    type: Literal["WEBHOOK"]
    app_host: AnyHttpUrl = Field(..., sensitive=True)

    def to_request(self) -> dict[str, Any]:
        return {
            **super().to_request(),
            "url": self.app_host + "/resync",
        }


class HttpEventListener(BaseEventListener):
    """
    HTTP event listener that listens for webhook events and triggers "on_resync" event.

    This class inherits from `BaseEventListener`, which provides a foundation for creating event listeners.
    The `HttpEventListener` listens for HTTP POST requests to the `/resync` endpoint and triggers the "on_resync" event.

    Parameters:
        events (EventListenerEvents): A dictionary containing event types and their corresponding event handlers.
        event_listener_config (HttpEventListenerSettings): Configuration settings for the HTTP event listener.
    """

    def __init__(
        self,
        events: EventListenerEvents,
        event_listener_config: HttpEventListenerSettings,
    ):
        super().__init__(events)
        self.event_listener_config = event_listener_config

    async def _start(self) -> None:
        """
        Starts the HTTP event listener.
        It sets up an APIRouter to handle the `/resync` endpoint and registers the "on_resync" event handler.
        """
        logger.info("Setting up HTTP Event Listener")
        target_channel_router = APIRouter()

        @target_channel_router.post("/resync")
        async def resync() -> None:
            await self._resync({})

        ocean.app.fast_api_app.include_router(target_channel_router)



from .events import EventsMixin
from .handler import HandlerMixin
from .sync import SyncMixin
from .sync_raw import SyncRawMixin

__all__ = [
    "EventsMixin",
    "HandlerMixin",
    "SyncRawMixin",
    "SyncMixin",
]



from typing import TypedDict, Any, AsyncIterator, Callable, Awaitable, NamedTuple

from port_ocean.core.models import Entity

RAW_ITEM = dict[Any, Any]
RAW_RESULT = list[RAW_ITEM]
ASYNC_GENERATOR_RESYNC_TYPE = AsyncIterator[RAW_RESULT]
RESYNC_RESULT = list[RAW_ITEM | ASYNC_GENERATOR_RESYNC_TYPE]

LISTENER_RESULT = Awaitable[RAW_RESULT] | ASYNC_GENERATOR_RESYNC_TYPE
RESYNC_EVENT_LISTENER = Callable[[str], LISTENER_RESULT]
START_EVENT_LISTENER = Callable[[], Awaitable[None]]


class RawEntityDiff(TypedDict):
    before: list[RAW_ITEM]
    after: list[RAW_ITEM]


class EntityDiff(TypedDict):
    before: list[Entity]
    after: list[Entity]


class EntitySelectorDiff(NamedTuple):
    passed: list[Entity]
    failed: list[Entity]


class CalculationResult(NamedTuple):
    entity_selector_diff: EntitySelectorDiff
    errors: list[Exception]


class IntegrationEventsCallbacks(TypedDict):
    start: list[START_EVENT_LISTENER]
    resync: dict[str | None, list[RESYNC_EVENT_LISTENER]]



# -*- coding: utf-8 -*-

import json
import os

import click
from cookiecutter.main import cookiecutter  # type: ignore

from port_ocean.cli.commands.main import cli_start, console, print_logo
from port_ocean.cli.utils import cli_root_path


def add_vscode_configuration(result: str, name: str) -> None:
    vscode_entry_root_path = "${workspaceFolder}/integrations/" + name
    new_vscode_entry = {
        "console": "integratedTerminal",
        "cwd": vscode_entry_root_path,
        "envFile": f"{vscode_entry_root_path}/.env",
        "justMyCode": True,
        "name": f"Run {name} integration",
        "program": f"{vscode_entry_root_path}/debug.py",
        "python": f"{vscode_entry_root_path}/.venv/bin/python",
        "request": "launch",
        "type": "debugpy",
    }

    vs_code_json_path = os.path.join(os.path.dirname(result), "../.vscode/launch.json")
    if not os.path.exists(vs_code_json_path):
        return
    vs_code_json = json.load(open(vs_code_json_path, "r"))
    vs_code_json["configurations"].append(new_vscode_entry)

    with open(vs_code_json_path, "w") as vs_code_json_file:
        json.dump(vs_code_json, vs_code_json_file, indent=2)
        vs_code_json_file.write("\n")


@cli_start.command()
@click.argument("path", default=".", type=click.Path(exists=True))
@click.option(
    "--private/--public",
    "is_private_integration",
    is_flag=True,
    default=True,
    help="Handle private integrations scaffolding. (Used for contributing to Ocean when set to public)",
)
def new(path: str, is_private_integration: bool) -> None:
    """
    Scaffold a new integration in the given PATH.

    PATH: Path to the integration. If not provided, the current directory will be used.
    """
    print_logo()

    console.print(
        "🚢 Unloading cargo... Setting up your integration at the dock.", style="bold"
    )

    result = cookiecutter(
        f"{cli_root_path}/cookiecutter",
        output_dir=path,
        extra_context={
            "is_private_integration": is_private_integration,
        },
    )

    name = result.split("/")[-1]

    final_private_integration = os.path.exists(os.path.join(result, "Dockerfile"))

    if not final_private_integration:
        add_vscode_configuration(result, name)

    console.print(
        "\n🌊 Ahoy, Captain! Your project is ready to set sail into the vast ocean of possibilities!",
        style="bold",
    )
    console.print("Here are your next steps:\n", style="bold")
    console.print(
        "⚓️ Install necessary packages: Run [bold][blue]make install[/blue][/bold] to install all required packages for your project.\n"
        f"▶️ [bold][blue]cd {path}/{name} && make install && . .venv/bin/activate[/blue][/bold]\n"
    )
    console.print(
        f"⚓️ Copy example env file: Run [bold][blue]cp {path}/{name}.env.example {path}/{name}/.env [/blue][/bold] and set your port credentials in the created file.\n"
    )
    console.print(
        "⚓️ Set sail with [blue]Ocean[/blue]: Run [bold][blue]ocean sail[/blue] <path_to_integration>[/bold] to run the project using Ocean.\n"
        f"▶️ [bold][blue]ocean sail {path}/{name}[/blue][/bold] \n"
    )
    if not final_private_integration:
        console.print(
            "⚓️ Smooth sailing with [blue]Make[/blue]: Alternatively, you can run [bold][blue]make run[/blue][/bold] to launch your project using Make. \n"
            f"▶️ [bold][blue]make run {path}/{name}[/blue][/bold]"
        )



import re
from typing import Any, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from loguru import Record

# https://github.com/h33tlit/secret-regex-list
secret_patterns = {
    "Password in URL": r"[a-zA-Z]{3,10}:\/\/[^\/\\s:@]{3,20}:[^\/\\s:@]{3,20}@.{1,100}[\"'\\s]",
    "Generic API Key": r"[a|A][p|P][i|I][_]?[k|K][e|E][y|Y].*['|\"][0-9a-zA-Z]{32,45}['|\"]",
    "Generic Secret": r"[s|S][e|E][c|C][r|R][e|E][t|T].*['|\"][0-9a-zA-Z]{32,45}['|\"]",
    "Google API Key": r"AIza[0-9A-Za-z\\-_]{35}",
    "Firebase URL": r".*firebaseio\.com",
    "RSA private key": r"-----BEGIN RSA PRIVATE KEY-----",
    "SSH (DSA) private key": r"-----BEGIN DSA PRIVATE KEY-----",
    "SSH (EC) private key": r"-----BEGIN EC PRIVATE KEY-----",
    "PGP private key block": r"-----BEGIN PGP PRIVATE KEY BLOCK-----",
    "Amazon AWS Access Key ID": r"AKIA[0-9A-Z]{16}",
    "Amazon MWS Auth Token": r"amzn\\.mws\\.[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}",
    "AWS API Key": r"AKIA[0-9A-Z]{16}",
    "GitHub": r"[g|G][i|I][t|T][h|H][u|U][b|B].*['|\"][0-9a-zA-Z]{35,40}['|\"]",
    "Google Cloud Platform API Key": r"AIza[0-9A-Za-z\\-_]{35}",
    "Google Cloud Platform OAuth": r"[0-9]+-[0-9A-Za-z_]{32}\\.apps\\.googleusercontent\\.com",
    "Google (GCP) Service-account": f'"type":{" "}"service_account"',
    "Google OAuth Access Token": r"ya29\\.[0-9A-Za-z\\-_]+",
    "Connection String": r"[a-zA-Z]+:\/\/[^/\s]+:[^/\s]+@[^/\s]+\/[^/\s]+",
}


class SensitiveLogFilter:
    compiled_patterns = [re.compile(pattern) for pattern in secret_patterns.values()]

    def hide_sensitive_strings(self, *tokens: str) -> None:
        self.compiled_patterns.extend(
            [re.compile(re.escape(token.strip())) for token in tokens if token.strip()]
        )

    def mask_string(self, string: str, full_hide: bool = False) -> str:
        masked_string = string
        for pattern in self.compiled_patterns:
            replace: Callable[[re.Match[str]], str] | str = (
                "[REDACTED]"
                if full_hide
                else lambda match: match.group()[:6] + "[REDACTED]"
            )
            masked_string = pattern.sub(replace, masked_string)
        return masked_string

    def mask_object(self, obj: Any, full_hide: bool = False) -> Any:
        if isinstance(obj, str):
            return self.mask_string(obj, full_hide)
        if isinstance(obj, list):
            return [self.mask_object(o, full_hide) for o in obj]
        if isinstance(obj, dict):
            for k, v in obj.items():
                obj[k] = self.mask_object(v, full_hide)

        return obj

    def create_filter(self, full_hide: bool = False) -> Callable[["Record"], bool]:
        def _filter(record: "Record") -> bool:
            record["message"] = self.mask_string(record["message"], full_hide)
            return True

        return _filter


sensitive_log_filter = SensitiveLogFilter()



from typing import Any

import httpx
from loguru import logger
from pydantic import BaseModel, Field, PrivateAttr

from port_ocean.clients.port.types import UserAgentType
from port_ocean.clients.port.utils import handle_status_code
from port_ocean.utils.misc import get_time


class TokenResponse(BaseModel):
    access_token: str = Field(alias="accessToken")
    expires_in: int = Field(alias="expiresIn")
    token_type: str = Field(alias="tokenType")
    _retrieved_time: int = PrivateAttr(default_factory=lambda: int(get_time()))

    @property
    def expired(self) -> bool:
        return self._retrieved_time + self.expires_in < get_time()

    @property
    def full_token(self) -> str:
        return f"{self.token_type} {self.access_token}"


class PortAuthentication:
    def __init__(
        self,
        client: httpx.AsyncClient,
        client_id: str,
        client_secret: str,
        api_url: str,
        integration_identifier: str,
        integration_type: str,
        integration_version: str,
    ):
        self.client = client
        self.api_url = api_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.integration_identifier = integration_identifier
        self.integration_type = integration_type
        self.integration_version = integration_version
        self.last_token_object: TokenResponse | None = None

    async def _get_token(self, client_id: str, client_secret: str) -> TokenResponse:
        logger.info(f"Fetching access token for clientId: {client_id}")

        credentials = {"clientId": client_id, "clientSecret": client_secret}
        response = await self.client.post(
            f"{self.api_url}/auth/access_token",
            json=credentials,
            extensions={"retryable": True},
        )
        handle_status_code(response)
        return TokenResponse(**response.json())

    def user_agent(self, user_agent_type: UserAgentType | None = None) -> str:
        user_agent = f"port-ocean/{self.integration_type}/{self.integration_version}/{self.integration_identifier}"
        if user_agent_type:
            user_agent += f"/{user_agent_type.value or UserAgentType.exporter.value}"

        return user_agent

    async def headers(
        self, user_agent_type: UserAgentType | None = None
    ) -> dict[Any, Any]:
        return {
            "Authorization": await self.token,
            "User-Agent": self.user_agent(user_agent_type),
        }

    @property
    async def token(self) -> str:
        if not self.last_token_object or self.last_token_object.expired:
            msg = "Token expired, fetching new token"
            if not self.last_token_object:
                msg = "No token found, fetching new token"
            logger.info(msg)
            self.last_token_object = await self._get_token(
                self.client_id, self.client_secret
            )
        return self.last_token_object.full_token



from typing import Any, Callable, Type

import httpx
from loguru import logger

from port_ocean.helpers.retry import RetryTransport


class OceanAsyncClient(httpx.AsyncClient):
    """
    This class is a wrapper around httpx.AsyncClient that uses a custom transport class.
    This is done to allow passing our custom transport class to the AsyncClient constructor while still allowing
    all the default AsyncClient behavior that is changed when passing a custom transport instance.
    """

    def __init__(
        self,
        transport_class: Type[RetryTransport] = RetryTransport,
        transport_kwargs: dict[str, Any] | None = None,
        **kwargs: Any,
    ):
        self._transport_kwargs = transport_kwargs
        self._transport_class = transport_class
        super().__init__(**kwargs)

    def _init_transport(  # type: ignore[override]
        self,
        transport: httpx.AsyncBaseTransport | None = None,
        app: Callable[..., Any] | None = None,
        **kwargs: Any,
    ) -> httpx.AsyncBaseTransport:
        if transport is not None or app is not None:
            return super()._init_transport(transport=transport, app=app, **kwargs)

        return self._transport_class(
            wrapped_transport=httpx.AsyncHTTPTransport(
                **kwargs,
            ),
            logger=logger,
            **(self._transport_kwargs or {}),
        )

    def _init_proxy_transport(  # type: ignore[override]
        self, proxy: httpx.Proxy, **kwargs: Any
    ) -> httpx.AsyncBaseTransport:
        return self._transport_class(
            wrapped_transport=httpx.AsyncHTTPTransport(
                proxy=proxy,
                **kwargs,
            ),
            logger=logger,
            **(self._transport_kwargs or {}),
        )



import sys
from inspect import getmembers
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

from yaml import safe_load

from port_ocean.bootstrap import create_default_app
from port_ocean.config.dynamic import default_config_factory
from port_ocean.core.handlers.port_app_config.models import ResourceConfig
from port_ocean.core.ocean_types import RESYNC_RESULT
from port_ocean.ocean import Ocean
from port_ocean.utils.misc import get_spec_file, load_module


def get_integration_ocean_app(
    integration_path: str, config_overrides: Union[Dict[str, Any], None] = None
) -> Ocean:
    spec_file = get_spec_file(Path(integration_path))
    config_factory = None
    if spec_file is not None:
        config_factory = default_config_factory(spec_file.get("configurations", []))

    default_app = create_default_app(
        integration_path,
        config_factory,
        {
            **(config_overrides or {}),
            **{
                "port": {
                    "client_id": "bla",
                    "client_secret": "bla",
                },
            },
        },
    )

    main_path = f"{integration_path}/main.py"
    sys.path.append(integration_path)
    app_module = load_module(main_path)
    app: Ocean = {name: item for name, item in getmembers(app_module)}.get(
        "app", default_app
    )

    return app


def get_integation_resource_configs(integration_path: str) -> List[ResourceConfig]:
    config_file_path = f"{integration_path}/.port/resources/port-app-config."
    if not Path(f"{config_file_path}yml").exists():
        config_file_path = f"{config_file_path}yaml"
    else:
        config_file_path = f"{config_file_path}yml"

    with open(config_file_path) as port_app_config_file:
        resource_configs = safe_load(port_app_config_file)

    return [ResourceConfig(**item) for item in resource_configs["resources"]]


async def get_raw_result_on_integration_sync_resource_config(
    app: Ocean, resource_config: ResourceConfig
) -> Tuple[RESYNC_RESULT, List[Exception]]:
    resource_result = await app.integration._get_resource_raw_results(resource_config)

    return resource_result



import asyncio
from inspect import getmembers
from typing import Dict, Any, Type

import uvicorn
from pydantic import BaseModel

from port_ocean.bootstrap import create_default_app
from port_ocean.config.dynamic import default_config_factory
from port_ocean.config.settings import ApplicationSettings, LogLevelType
from port_ocean.core.defaults.initialize import initialize_defaults
from port_ocean.core.utils import validate_integration_runtime
from port_ocean.log.logger_setup import setup_logger
from port_ocean.ocean import Ocean
from port_ocean.utils.misc import get_spec_file, load_module
from port_ocean.utils.signal import init_signal_handler


def _get_default_config_factory() -> None | Type[BaseModel]:
    spec = get_spec_file()
    config_factory = None
    if spec is not None:
        config_factory = default_config_factory(spec.get("configurations", []))

    return config_factory


def run(
    path: str = ".",
    log_level: LogLevelType = "INFO",
    port: int = 8000,
    initialize_port_resources: bool | None = None,
    config_override: Dict[str, Any] | None = None,
) -> None:
    application_settings = ApplicationSettings(log_level=log_level, port=port)

    init_signal_handler()
    setup_logger(
        application_settings.log_level,
        enable_http_handler=application_settings.enable_http_logging,
    )

    config_factory = _get_default_config_factory()
    default_app = create_default_app(path, config_factory, config_override)

    main_path = f"{path}/main.py" if path else "main.py"
    app_module = load_module(main_path)
    app: Ocean = {name: item for name, item in getmembers(app_module)}.get(
        "app", default_app
    )

    # Validate that the current integration's runtime matches the execution parameters
    asyncio.get_event_loop().run_until_complete(
        validate_integration_runtime(app.port_client, app.config.runtime)
    )

    # Override config with arguments
    if initialize_port_resources is not None:
        app.config.initialize_port_resources = initialize_port_resources

    initialize_defaults(app.integration.AppConfigHandlerClass.CONFIG_CLASS, app.config)

    uvicorn.run(app, host="0.0.0.0", port=application_settings.port)



from typing import Type

from loguru import logger

from port_ocean.context.ocean import ocean
from port_ocean.core.handlers import (
    BaseEntityProcessor,
    BasePortAppConfig,
    BaseEntitiesStateApplier,
    HttpEntitiesStateApplier,
    JQEntityProcessor,
    APIPortAppConfig,
)
from port_ocean.exceptions.core import IntegrationNotStartedException


class HandlerMixin:
    """Mixin class for handling integration components.

    This mixin class provides methods and properties for initializing and accessing various
    integration components, including the Entity Processor, PortAppConfig Handler, and
    EntitiesStateApplier.

    Attributes:
        EntityProcessorClass (Type[BaseEntityProcessor]): The class representing the Entity Processor.
        AppConfigHandlerClass (Type[BasePortAppConfig]): The class representing the PortAppConfig Handler.
        EntitiesStateApplierClass (Type[BaseEntitiesStateApplier]): The class representing the EntitiesStateApplier.
    """

    EntityProcessorClass: Type[BaseEntityProcessor] = JQEntityProcessor

    AppConfigHandlerClass: Type[BasePortAppConfig] = APIPortAppConfig

    EntitiesStateApplierClass: Type[BaseEntitiesStateApplier] = HttpEntitiesStateApplier

    def __init__(self) -> None:
        self._entity_processor: BaseEntityProcessor | None = None
        self._port_app_config_handler: BasePortAppConfig | None = None
        self._entities_state_applier: BaseEntitiesStateApplier | None = None

    @property
    def entity_processor(self) -> BaseEntityProcessor:
        """Access the Entity Processor instance.

        Raises:
            IntegrationNotStartedException: If the Entity Processor class is not initialized.
        """
        if not self._entity_processor:
            raise IntegrationNotStartedException(
                "Entity Processor class not initialized"
            )
        return self._entity_processor

    @property
    def port_app_config_handler(self) -> BasePortAppConfig:
        """Access the PortAppConfig Handler instance.

        Raises:
            IntegrationNotStartedException: If the PortAppConfig Handler class is not initialized.
        """
        if self._port_app_config_handler is None:
            raise IntegrationNotStartedException("PortAppConfig class not initialized")
        return self._port_app_config_handler

    @property
    def entities_state_applier(self) -> BaseEntitiesStateApplier:
        """Access the EntitiesStateApplier instance.

        Raises:
            IntegrationNotStartedException: If the EntitiesStateApplier class is not initialized.
        """
        if not self._entities_state_applier:
            raise IntegrationNotStartedException(
                "EntitiesStateApplier class not initialized"
            )
        return self._entities_state_applier

    async def _init_entity_processor_instance(self) -> BaseEntityProcessor:
        self._entity_processor = self.EntityProcessorClass(ocean)
        return self._entity_processor

    async def _init_port_app_config_handler_instance(
        self,
    ) -> BasePortAppConfig:
        self._port_app_config_handler = self.AppConfigHandlerClass(ocean)
        return self._port_app_config_handler

    async def _init_entities_state_applier_instance(self) -> BaseEntitiesStateApplier:
        self._entities_state_applier = self.EntitiesStateApplierClass(ocean)
        return self._entities_state_applier

    async def initialize_handlers(self) -> None:
        """Initialize all integration components."""
        logger.info("Initializing integration components")
        await self._init_entity_processor_instance()
        await self._init_port_app_config_handler_instance()
        await self._init_entities_state_applier_instance()



from port_ocean.exceptions.base import BaseOceanException


class AbortDefaultCreationError(BaseOceanException):
    def __init__(
        self,
        blueprints_to_rollback: list[str],
        errors: list[Exception],
    ):
        self.blueprints_to_rollback = blueprints_to_rollback
        self.errors = errors
        super().__init__("Aborting defaults creation")


class UnsupportedDefaultFileType(BaseOceanException):
    pass



import datetime
from loguru import logger


def convert_str_to_utc_datetime(time_str: str) -> datetime.datetime | None:
    """
    Convert a string representing time to a datetime object.
    :param time_str: a string representing time in the format "2021-09-01T12:00:00Z"
    """
    aware_date = datetime.datetime.fromisoformat(time_str)
    if time_str.endswith("Z"):
        aware_date = datetime.datetime.fromisoformat(time_str.replace("Z", "+00:00"))
    return aware_date.astimezone(datetime.timezone.utc)


def convert_to_minutes(s: str) -> int:
    minutes_per_unit = {"s": 1 / 60, "m": 1, "h": 60, "d": 1440, "w": 10080}
    try:
        return int(int(s[:-1]) * minutes_per_unit[s[-1]])
    except Exception:
        logger.error(f"Failed converting string to minutes, {s}")
        raise ValueError(
            f"Invalid format. Expected a string ending with {minutes_per_unit.keys()}"
        )


def get_next_occurrence(
    interval_seconds: int,
    start_time: datetime.datetime,
    now: datetime.datetime | None = None,
) -> datetime.datetime:
    """
    Predict the next occurrence of an event based on interval, start time, and current time.

    :param interval_minutes: Interval between occurrences in minutes.
    :param start_time: Start time of the event as a datetime object.
    :param now: Current time as a datetime object.
    :return: The next occurrence time as a datetime object.
    """

    if now is None:
        now = datetime.datetime.now(tz=datetime.timezone.utc)
    # Calculate the total seconds elapsed since the start time
    elapsed_seconds = (now - start_time).total_seconds()

    # Calculate the number of intervals that have passed
    intervals_passed = int(elapsed_seconds // interval_seconds)

    # Calculate the next occurrence time
    next_occurrence = start_time + datetime.timedelta(
        seconds=(intervals_passed + 1) * interval_seconds
    )

    return next_occurrence



from .entities_state_applier.base import (
    BaseEntitiesStateApplier,
)
from .entities_state_applier.port.applier import (
    HttpEntitiesStateApplier,
)
from .entity_processor.base import BaseEntityProcessor
from .entity_processor.jq_entity_processor import (
    JQEntityProcessor,
)
from .port_app_config.api import APIPortAppConfig
from .port_app_config.base import BasePortAppConfig

__all__ = [
    "BaseEntityProcessor",
    "JQEntityProcessor",
    "BasePortAppConfig",
    "APIPortAppConfig",
    "BaseEntitiesStateApplier",
    "HttpEntitiesStateApplier",
]



from typing import TYPE_CHECKING

import httpx
from loguru import logger
from werkzeug.local import LocalStack, LocalProxy

from port_ocean.clients.port.retry_transport import TokenRetryTransport
from port_ocean.helpers.async_client import OceanAsyncClient

if TYPE_CHECKING:
    from port_ocean.clients.port.client import PortClient

# In case the framework sends more requests to port in parallel than allowed by the limits, a PoolTimeout exception will
# be raised.
# Raising defaults for the timeout, in addition to the limits, will allow request to wait for a connection for a longer
# period of time, before raising an exception.
# The max_connections value can't be too high, as it will cause the application to run out of memory.
# The max_keepalive_connections can't be too high, as it will cause the application to run out of available connections.
PORT_HTTP_MAX_CONNECTIONS_LIMIT = 200
PORT_HTTP_MAX_KEEP_ALIVE_CONNECTIONS = 50
PORT_HTTP_TIMEOUT = 60.0

PORT_HTTPX_TIMEOUT = httpx.Timeout(PORT_HTTP_TIMEOUT)
PORT_HTTPX_LIMITS = httpx.Limits(
    max_connections=PORT_HTTP_MAX_CONNECTIONS_LIMIT,
    max_keepalive_connections=PORT_HTTP_MAX_KEEP_ALIVE_CONNECTIONS,
)

_http_client: LocalStack[httpx.AsyncClient] = LocalStack()


def _get_http_client_context(port_client: "PortClient") -> httpx.AsyncClient:
    client = _http_client.top
    if client is None:
        client = OceanAsyncClient(
            TokenRetryTransport,
            transport_kwargs={"port_client": port_client},
            timeout=PORT_HTTPX_TIMEOUT,
            limits=PORT_HTTPX_LIMITS,
        )
        _http_client.push(client)

    return client


_port_internal_async_client: httpx.AsyncClient = None  # type: ignore


def get_internal_http_client(port_client: "PortClient") -> httpx.AsyncClient:
    global _port_internal_async_client
    if _port_internal_async_client is None:
        _port_internal_async_client = LocalProxy(
            lambda: _get_http_client_context(port_client)
        )

    return _port_internal_async_client


def handle_status_code(
    response: httpx.Response, should_raise: bool = True, should_log: bool = True
) -> None:
    if should_log and response.is_error:
        logger.error(
            f"Request failed with status code: {response.status_code}, Error: {response.text}"
        )
    if should_raise:
        response.raise_for_status()



import sys
from logging import LogRecord
from logging.handlers import QueueHandler, QueueListener
from queue import Queue

import loguru
from loguru import logger

from port_ocean.config.settings import LogLevelType
from port_ocean.log.handlers import HTTPMemoryHandler
from port_ocean.log.sensetive import sensitive_log_filter
from port_ocean.utils.signal import signal_handler


def setup_logger(level: LogLevelType, enable_http_handler: bool) -> None:
    logger.remove()
    _stdout_loguru_handler(level)
    if enable_http_handler:
        _http_loguru_handler(level)


def _stdout_loguru_handler(level: LogLevelType) -> None:
    logger_format = (
        "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
        "<level>{level: <8}</level> | "
        "<level>{message}</level>"
    )
    if level == "DEBUG":
        logger_format += " | {extra}"

    logger.add(
        sys.stdout,
        level=level.upper(),
        format=logger_format,
        enqueue=True,  # process logs in background
        diagnose=False,  # hide variable values in log backtrace
        filter=sensitive_log_filter.create_filter(),
    )
    logger.configure(patcher=exception_deserializer)


def _http_loguru_handler(level: LogLevelType) -> None:
    queue: Queue[LogRecord] = Queue()

    handler = QueueHandler(queue)

    logger.add(
        handler,
        level=level.upper(),
        format="{message}",
        diagnose=False,  # hide variable values in log backtrace
        enqueue=True,  # process logs in background
        filter=sensitive_log_filter.create_filter(full_hide=True),
    )
    logger.configure(patcher=exception_deserializer)

    http_memory_handler = HTTPMemoryHandler()
    signal_handler.register(http_memory_handler.wait_for_lingering_threads)
    signal_handler.register(http_memory_handler.flush)

    queue_listener = QueueListener(queue, http_memory_handler)
    queue_listener.start()


def exception_deserializer(record: "loguru.Record") -> None:
    """
    Workaround for when trying to log exception objects with loguru.
    Loguru doesn't able to deserialize `Exception` subclasses.
    https://github.com/Delgan/loguru/issues/504#issuecomment-917365972
    """
    exception: loguru.RecordException | None = record["exception"]
    if exception is not None:
        fixed = Exception(str(exception.value))
        record["exception"] = exception._replace(value=fixed)



import httpx
from werkzeug.local import LocalStack, LocalProxy

from port_ocean.context.ocean import ocean
from port_ocean.helpers.async_client import OceanAsyncClient
from port_ocean.helpers.retry import RetryTransport

_http_client: LocalStack[httpx.AsyncClient] = LocalStack()


def _get_http_client_context() -> httpx.AsyncClient:
    client = _http_client.top
    if client is None:
        client = OceanAsyncClient(RetryTransport, timeout=ocean.config.client_timeout)
        _http_client.push(client)

    return client


"""
Utilize this client for all outbound integration requests to the third-party application. It functions as a wrapper 
around the httpx.AsyncClient, incorporating retry logic at the transport layer for handling retries on 5xx errors and
connection errors.

The client is instantiated lazily, only coming into existence upon its initial access. It should not be closed when in
use, as it operates as a singleton shared across all events in the thread. It also takes care of recreating the client
in scenarios such as the creation of a new event loop, such as when initiating a new thread.
"""
http_async_client: httpx.AsyncClient = LocalProxy(lambda: _get_http_client_context())  # type: ignore
